"""Cross-space module
"""

import numpy as np
import pandas as pd


def cross_proximity(rcas_a, rcas_b):
    """Calculates the Cross-proximity index between two matrices of RCA.

    Catalan et al. (2020) introduced the cross-proximity index in order
    to measure the minimum probability that a country presents comparative
    advantages in a patent in a given area, given it has comparative
    advantages in an area of ​​knowledge, or vice versa, where values ​​that
    are more close to unity indicate a stronger relationship between the
    patent area and the knowledge area.

    Note the characteristic in both RCA matrices can't be the location.

    Args:
        rcas_a (pd.DataFrame) -- A matrix of pivotted data for RCA, obtained
            from the RCA function, for a main characteristic to evaluate.
        rcas_b (pd.DataFrame) -- The pivot table obtained from the RCA
            function, for a secondary characteristic to evaluate.

    Returns:
        (pd.DataFrame) -- A matrix with the proximity between the two types
            of evaluated elements that can be used in the calculation of the
            cross-relatedness.
    """
    # Trims RCA values to [0,1]
    rcas_a[rcas_a >= 1] = 1
    rcas_a[rcas_a < 1] = 0
    rcas_b[rcas_b >= 1] = 1
    rcas_b[rcas_b < 1] = 0

    # Calculates numerator of array
    numerator = rcas_a.T.dot(rcas_b)

    # Calculates kp0 for rcas_a and rcas_b
    kp0_a = rcas_a.sum(axis=0)
    kp0_a = kp0_a.values.reshape((1, len(kp0_a)))

    kp0_b = rcas_b.sum(axis=0)
    kp0_b = kp0_b.values.reshape((1, len(kp0_b)))

    # Calculates two possible cross proximity values and replace posibles NaN values
    a = numerator.divide(kp0_b).fillna(0)
    b = numerator.divide(kp0_a.T).fillna(0)

    # Compares the two previous arrays, and keeps minimum value of each cell
    x_proximity = pd.DataFrame(
        [np.minimum(x, y) for x, y in zip(a.values, b.values)],
        index=a.index,
        columns=a.columns
    )
    return x_proximity


def cross_relatedness(rcas: pd.DataFrame, x_proximity: pd.DataFrame):
    """Calculates the Cross-relatedness.

    Catalan et al. (2020) incorporated the concept of cross-relatedness
    in order to quantify the relationship between scientific fields in the
    development of new technologies in a country. The density crossing
    takes values ​​between zero and one, mathematically it corresponds to
    the average cross proximity of a technology and the scientific knowledge
    of a country during the period of time.

    Note it's important to be consistent with the name of the variables
    when calculating cross-proximity and cross-relatedness.
    To display the outputted values, the series must be transformed into
    a dataframe in tidy format as shown below:

        x_relatedness = cross_relatedness(rcas, cross_proximity)
        x_relatedness.unstack().reset_index(name="cross_relatedness")

    Args:
        rcas (pd.DataFrame) -- A pivotted table obtained from the RCA
            function, for the main characteristic to evaluate.
        x_proximity {} -- The cross-proximity matrix obtained between the
            `rcas` matrix in the first parameter, and another RCA matrix.

    Returns:
        (pd.DataFrame) -- A matrix with the probability that a location
            generates comparative advantages in the characteristic to be
            evaluated considering its proximity with the other evaluated
            characteristic.
    """
    rcas = rcas.copy()
    rcas[rcas >= 1] = 1
    rcas[rcas < 1] = 0

    numerator = rcas.dot(x_proximity)

    rcas_ones = rcas * 0
    rcas_ones = rcas + 1

    denominator = rcas_ones.dot(x_proximity)

    x_relatedness = numerator / denominator
    return x_relatedness



"""Economic Complexity

This module contains functions to ease the calculation of Economic Complexity values.
"""

from .rca import rca
from .complexity import complexity
from .product_space import distance, opportunity_gain, proximity, relatedness, similarity, pgi, peii
from .cross_space import cross_proximity, cross_relatedness

__version_info__ = ('0', '1', '2')
__version__ = '.'.join(__version_info__)

__all__ = (
    "rca",
    "complexity",
    "distance",
    "opportunity_gain",
    "proximity",
    "relatedness",
    "cross_proximity",
    "cross_relatedness",
    "similarity",
    "pgi",
    "peii",
)



"""Complexity Indices module

Hidalgo & Hausmann (2009), calculate the Economic Complexity Indices from the
reflections method, which is defined based on a red bipartite that contains a
symmetric set of variables whose nodes correspond to countries and products.
"""

import logging
from typing import Optional, Tuple

import pandas as pd

logger = logging.getLogger(__name__)


def complexity(rca: pd.DataFrame,
               iterations: int = 20,
               drop: Optional[bool] = True) -> Tuple[pd.Series, pd.Series]:
    """Calculates Economic Complexity Index (ECI) and Product Complexity
    Index (PCI) from a RCA matrix.

    Note that to display the resulting values, the series must be transformed
    into a dataframe in an tidy format as shown below:

        eci_value, pci_value = complexity(rca)
        eci = eci_value.to_frame(name="ECI").reset_index()
        pci = pci_value.to_frame(name="PCI").reset_index()

    Args:
        rcas (pd.DataFrame) -- Pivotted RCA matrix.
        iterations (int, optional) -- Limit of recursive calculations for
            kp and kc. Default value: 20.
        drop (bool, optional) -- Boolean to ensure that returns include NaN
            values. Default value: True.

    Returns:
        ((pd.Series, pd.Series)) -- A tuple of ECI and PCI values.
    """
    # Binarize rca input
    rcas = pd.DataFrame(rca.copy())
    rcas[rcas >= 1] = 1
    rcas[rcas < 1] = 0

    # drop columns / rows only if completely nan
    rcas_clone = rcas.copy()
    rcas_clone = rcas_clone.dropna(how="all")
    rcas_clone = rcas_clone.dropna(how="all", axis=1)

    if rcas_clone.shape != rcas.shape:
        logger.warning("RCAs contain columns or rows that are entirely comprised of NaN values.")
    if drop:
        rcas = rcas_clone

    kp = rcas.sum(axis=0) #sum columns
    kc = rcas.sum(axis=1) #sum rows
    kp0 = kp.copy()
    kc0 = kc.copy()

    for i in range(1, iterations):
        kc_temp = kc.copy()
        kp_temp = kp.copy()
        kp = rcas.T.dot(kc_temp) / kp0
        if i < (iterations - 1):
            kc = rcas.dot(kp_temp) / kc0

    geo_complexity = (kc - kc.mean()) / kc.std()
    prod_complexity = (kp - kp.mean()) / kp.std()

    return geo_complexity, prod_complexity



"""Revealed Comparative Advantage (RCA) module

The Revealed Comparative Advantage is an index introduced by Balassa (1965),
which is used to evaluate the main products to be exported by a country,
and their comparative advantages in relation to the level of world exports
(Hidalgo et al., 2007).
"""

import numpy as np
import pandas as pd


def rca(tbl: pd.DataFrame) -> pd.DataFrame:
    """Calculates the Revealed Comparative Advantage (RCA) for a pivoted matrix.

    It is important to note that even though the functions do not use a
    parameter in relation to time, the data used for the calculations must
    be per period; for example working with World Exports for the year 2020.
    Also, the index always has to be a geographic level.

    Arguments:
        tbl (pandas.DataFrame) -- A pivoted table using a geographic index,
            columns with the categories to be evaluated and the measurement of
            the data as values.

    Returns:
        (pandas.DataFrame) -- RCA matrix with real values.
    """
    # fill missing values with zeros
    tbl = tbl.fillna(value=0)

    col_sums = tbl.sum(axis=1) # 1:columns
    col_sums = col_sums.to_numpy().reshape((len(col_sums), 1))

    rca_numerator = np.divide(tbl, col_sums)
    row_sums = tbl.sum(axis=0) # 0:index

    total_sum = tbl.sum().sum()
    rca_denominator = row_sums / total_sum
    rcas = rca_numerator / rca_denominator

    return rcas



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['economic_complexity']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18.0,<2.0.0', 'pandas>=1.1.0,<2.0.0']

setup_kwargs = {
    'name': 'economic-complexity',
    'version': '0.1.2',
    'description': 'Functions to calculate Economic Complexity indicators.',
    'long_description': 'This package contains functions to calculate Economic Complexity indicators.  \nThe functions handle the data through `pandas.DataFrame` objects.\n\n<p>\n<a href="https://github.com/Datawheel/py-economic-complexity"><img src="https://flat.badgen.net/github/release/Datawheel/py-economic-complexity" /></a>\n<a href="https://github.com/Datawheel/py-economic-complexity/blob/master/LICENSE"><img src="https://flat.badgen.net/github/license/Datawheel/py-economic-complexity" /></a>\n<a href="https://github.com/Datawheel/py-economic-complexity/"><img src="https://flat.badgen.net/github/checks/Datawheel/py-economic-complexity" /></a>\n<a href="https://github.com/Datawheel/py-economic-complexity/issues"><img src="https://flat.badgen.net/github/issues/Datawheel/py-economic-complexity" /></a>\n</p>\n\n## Installation\n\nWe recommend the use of `poetry`, to resolve the best version of the dependencies that works with your current project.\n\n```bash\n$ poetry add economic-complexity\n```\n\n## Tutorial\n\nWe have [a brief Tutorial](https://github.com/Datawheel/py-economic-complexity/blob/main/docs/TUTORIAL.ipynb), using data from the Observatory of Economic Complexity, to get started on how to use the basic functions of this package.  \nMore complex functions use the resulting dataframes of the basic functions as arguments.\n\n## References\n\n* Hidalgo, César A. (2021). Economic complexity theory and applications. _Nature Reviews Physics, 3_(2), 92–113. https://doi.org/10.1038/s42254-020-00275-1\n\n* Catalán, P., Navarrete, C., & Figueroa, F. (2020). The scientific and technological cross-space: Is technological diversification driven by scientific endogenous capacity? _Research Policy, 104016_, 104016. https://doi.org/10.1016/j.respol.2020.104016\n\n* Hidalgo, César A., & Hausmann, R. (2009). The building blocks of economic complexity. _Proceedings of the National Academy of Sciences of the United States of America, 106_(26), 10570–10575. https://doi.org/10.1073/pnas.0900943106\n\n* Hidalgo, C. A., Klinger, B., Barabási, A.-L., & Hausmann, R. (2007). The product space conditions the development of nations. _Science (New York, N.Y.), 317_(5837), 482–487. https://doi.org/10.1126/science.1144581\n\n* Hartmann, D., Guevara, M. R., Jara-Figueroa, C., Aristarán, M., & Hidalgo, C. A. (2017). Linking Economic Complexity, Institutions, and Income Inequality. _World Development_, 93, 75–93. https://doi.org/10.1016/j.worlddev.2016.12.020\n\n* Romero, J. P., & Gramkow, C. (2021). Economic complexity and greenhouse gas emissions. _World Development_, 139, 105317. https://doi.org/10.1016/j.worlddev.2020.105317\n\n* Bahar, D., Hausmann, R., Hidalgo, C. A. (2014). Neighbors and the evolution of the comparative advantage of nations: Evidence of international knowledge diffusion?. _Journal of International Economics_, 92, 111-123. http://dx.doi.org/10.1016/j.jinteco.2013.11.001\n\n---\n&copy; 2022 [Datawheel, LLC.](https://www.datawheel.us/)  \nThis project is licensed under [MIT](./LICENSE).\n',
    'author': 'Jelmy Hermosilla',
    'author_email': 'jelmy@datawheel.us',
    'maintainer': 'Francisco Abarzua',
    'maintainer_email': 'francisco@datawheel.us',
    'url': 'https://github.com/Datawheel/py-economic-complexity',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Product Space module
"""

import pandas as pd
import numpy as np


def proximity(rcas: pd.DataFrame, procedure="max"):
    """Calculates the Proximity index for a matrix of RCAs.

    Hidalgo et al. (2007), introduces the Proximity index which measures the
    minimum probability that a country has comparative advantages in the export
    of a product `i`, given that it has comparative advantages in a product `j`.

    This function only needs the pivot table obtained from the RCA function,
    and returns a square matrix with the proximity between the elements.

    Args:
        rcas (pd.DataFrame) -- A RCA matrix of pivotted values.
        procedure (str, optional) -- Determines how to calcule the denominator.
            Available options are "sqrt" and "max", defaults to "max".

    Returns:
        (pd.DataFrame) -- A square matrix with the proximity between the elements.
    """
    rcas = rcas.copy()
    rcas[rcas >= 1] = 1
    rcas[rcas < 1] = 0

    # transpose the matrix so that it is now industries as rows
    # and munics as columns
    rcas_t = rcas.T.fillna(0)

    # Matrix multiplication on M_mi matrix and transposed version,
    # number of products = number of rows and vice versa on transposed
    # version, thus the shape of this result will be length of products
    # by the length of products (symetric)
    numerator_intersection = rcas_t.dot(rcas_t.T)

    # kp0 is a vector of the number of munics with RCA in the given product
    kp0 = rcas.sum(axis=0)
    kp0 = kp0.to_numpy().reshape((1, len(kp0)))

    # transpose this to get the unions
    kp0_trans = kp0.T

    # multiply these two vectors, take the squre root
    # and then we have the denominator
    if procedure == "sqrt":
        # get square root for geometric mean
        denominator_union = kp0_trans.dot(kp0)
        denominator_union = np.power(denominator_union, .5)
    else:
        denominator_union = np.maximum(kp0, kp0_trans)

    # to get the proximities it is now a simple division of the untion sqrt
    # with the numerator intersections
    phi = np.divide(numerator_intersection, denominator_union)
    np.fill_diagonal(phi.values, 0)

    return phi


def relatedness(rcas: pd.DataFrame, proximities: pd.DataFrame):
    """Calculates the Relatedness, given a matrix of RCAs for the economic
    activities of a location, and a matrix of Proximities.

    The fact that the growth of an activity in a location is correlated with
    relatedness is known as The Principle of Relatedness, introduced by
    Hidalgo et al. (2018), which consists of a statistical law that tells us
    that the probability a location (a country, city, or region) enters an
    economic activity (e.g. a product, industry, technology), grows with the
    number of related activities present in that location.

    Scholars measure Relatedness between a location and an economic activity
    to predict the probability that a region will enter or exit that activity
    in the future.

    Args:
        rcas (pd.DataFrame) -- Matrix of RCAs for a certain location.
        proximities (pd.DataFrame) -- Matrix with the proximity between the elements.

    Returns:
        (pd.DataFrame) -- A matrix with the probability that a location
            generates comparative advantages in a economic activity.
    """
    rcas = rcas.copy()
    rcas[rcas >= 1] = 1
    rcas[rcas < 1] = 0

    # Get numerator by matrix multiplication of proximities with M_im
    density_numerator = rcas.dot(proximities)

    # Get denominator by multiplying proximities by all ones vector thus
    # getting the sum of all proximities
    # rcas_ones = pd.DataFrame(np.ones_like(rcas))
    rcas_ones = rcas * 0
    rcas_ones = rcas_ones + 1
    # print rcas_ones.shape, proximities.shape
    density_denominator = rcas_ones.dot(proximities)

    # We now have our densities matrix by dividing numerator by denomiator
    densities = density_numerator / density_denominator

    return densities


def distance(rcas: pd.DataFrame, proximities: pd.DataFrame):
    """Calculates the distance.

    In this context, the Distance for a Country and a certain Product is defined
    as the

    Args:
        rcas (pd.DataFrame) -- Matrix of RCAs for a certain location.
        proximities (pd.DataFrame) -- Matrix with the proximity between the elements.

    Returns:
        (pd.DataFrame) --
    """
    return 1 - relatedness(rcas, proximities)


def opportunity_gain(rcas: pd.DataFrame, proximities: pd.DataFrame, pci: pd.DataFrame):
    """Calculates the opportunity gain caused by the contribution of a certain
    characteristic, relative to how this affects other characteristics.

    Args:
        rcas (pd.DataFrame) -- Matrix of RCAs considering a
        proximities (pd.DataFrame) -- [description]
        pci (pd.DataFrame) -- [description]

    Returns:
        (pd.DataFrame) --
    """
    rcas = rcas.copy()
    rcas[rcas >= 1] = 1
    rcas[rcas < 1] = 0

    # turn proximities in to ratios out of total
    prox_ratio = proximities / proximities.sum()

    # get the inverse of the RCAs matrix. Since they are in the form of 1's and
    # 0's this will flip all of them i.e. 1 = 0 and 0 = 1
    inverse_rcas = 1 - rcas

    # here we now have the middle part of the equation
    middle = inverse_rcas.multiply(pci)

    # get the relatedness with the backwards bizzaro RCAs
    dcp = relatedness(inverse_rcas, proximities)
    # now get the inverse
    dcp = 1 - dcp
    # we now have the right-half of the equation
    right = dcp.multiply(pci)

    # matrix multiplication with proximities ratio
    opp_gain = middle.dot(prox_ratio) - right

    return opp_gain


def similarity(rcas: pd.DataFrame):
    """
    Calculates the Export Similarity Index for a matrix of RCAs.

    Bahar et al. (2014) introduces this measure of similarity in the export structure of a pair of countries c and c'. It's defined as the Pearson correlation between the logarithm of the RCA vectors of the two countries.

    This function only needs the pivot table obtained from the RCA function,
    and returns a square matrix with the Export Similarity Index between the elements.

    Args:
        rcas (pd.DataFrame) -- A RCA matrix of pivotted values.

    Returns:
        (pd.DataFrame) -- A square matrix with the Export Similarity Index between the elements.
    """

    # Take the log of rcas (add 0.1 as \epsilon)
    rcas = np.log(rcas + 0.1)

    # calculate the matrix through the pearson correlation
    scc = pd.DataFrame(np.corrcoef(rcas), columns=rcas.index, index = rcas.index)

    return scc


def pmi(tbl: pd.DataFrame, rcas: pd.DataFrame, measure: pd.DataFrame, measure_name: str) -> pd.DataFrame:
    """
    Calculates the Product 'measure' Index, where measure corresponds to a dataframe with the measure values for each geography.
    For example, in the literature this method has been applied to calculate the Product Gini Index (PGI) and the Product Emission Intensity Index.

    Args:
        tbl (pd.DataFrame) -- A pivoted table using a geographic index,
            columns with the categories to be evaluated and the measurement of
            the data as values.
        measure (pd.DataFrame) -- A table using a geographic index, with a single column with the measure values.
        measure_name (str) -- A string with the name of the measure

    Returns:
        (pd.DataFrame) -- A square matrix with the Export Similarity Index between the elements.

    """
    # drop product with no exports and fill missing values with zeros
    tbl = tbl.dropna(how="all", axis=1).fillna(value=0)
    measure = measure.fillna(value=0)

    # get Mcp matrix
    m = rcas.copy()
    m[rcas >= 1] = 1
    m[rcas < 1] = 0

    # Ensures that the matrices are aligned by removing geographies that don't exist in both matrices
    tbl_geo = tbl.index
    measure_geo = measure.index
    intersection_geo = list(set(tbl_geo) & set(measure_geo))
    tbl = tbl.filter(items=intersection_geo, axis=0)
    measure = measure.filter(items=intersection_geo, axis=0)
    m = m.filter(items=intersection_geo, axis=0)

    tbl = tbl.sort_index(ascending=True)
    measure = measure.sort_index(ascending=True)
    m = m.sort_index(ascending=True)

    # get Scp matrix
    col_sums = tbl.sum(axis=1)
    col_sums = col_sums.to_numpy().reshape((len(col_sums), 1))
    scp = np.divide(tbl, col_sums)

    # get Np array
    normp = m.multiply(scp).sum(axis=0)
    normp = pd.DataFrame(normp)

    num = m.multiply(scp).T.dot(measure)

    pmi = np.divide(num, normp)
    pmi.rename(columns={pmi.columns[0]: measure_name}, inplace=True)

    return pmi

def pgi(tbl: pd.DataFrame, rcas: pd.DataFrame, gini: pd.DataFrame) -> pd.DataFrame:

    """Calculates the Product Gini Index (PGI) for a pivoted matrix.
    It is important to note that even though the functions do not use a
    parameter in relation to time, the data used for the calculations must
    be per period; for example working with World Exports for the year 2020.
    Also, the index always has to be a geographic level.
    It is also important to make sure that the input matrices are aligned,
    that is, that both matrices consider the same geographic units.
    Arguments:
        tbl (pandas.DataFrame) -- A pivoted table using a geographic index,
            columns with the categories to be evaluated and the measurement of
            the data as values.
        gini (pandas.DataFrame) -- A matrix of GINI indices using a geographic index.
    Returns:
        (pandas.DataFrame) -- PGI matrix with categories evaluated as an index.
    """

    pgip = pmi(tbl = tbl, rcas = rcas, measure = gini, measure_name='pgi')

    return pgip

def peii(tbl: pd.DataFrame, rcas: pd.DataFrame, emissions: pd.DataFrame) -> pd.DataFrame:

    """
    Calculates the Product Emissions Intensity Index (PEII) for a pivoted matrix.
    It is important to note that even though the functions do not use a
    parameter in relation to time, the data used for the calculations must
    be per period; for example working with World Exports for the year 2020.
    Also, the index always has to be a geographic level.
    It is also important to make sure that the input matrices are aligned,
    that is, that both matrices consider the same geographic units.
    Arguments:
        tbl (pandas.DataFrame) -- A pivoted table using a geographic index,
            columns with the categories to be evaluated and the measurement of
            the data as values.
        emissions (pandas.DataFrame) -- A matrix of emissions intensity using a geographic index.
    Returns:
        (pandas.DataFrame) -- PEII matrix with categories evaluated as an index.
    """

    peii = pmi(tbl = tbl, rcas = rcas, measure=emissions, measure_name='peii')
    return peii
